"""migrate reaction dataset

Revision ID: 0b31ab4244de
Revises: 8a41b7dc30af
Create Date: 2022-03-15 12:51:52.995657

"""

import json
import os
import sys
from datetime import datetime, timezone

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcfractal import __version__ as qcfractal_version

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers.v0_50_helpers import get_empty_keywords_id, add_qc_spec

# revision identifiers, used by Alembic.
revision = "0b31ab4244de"
down_revision = "8a41b7dc30af"
branch_labels = None
depends_on = None


def create_reaction_record(conn, ds_id, created_on, modified_on, spec_id, stoichiometries, records):
    # We do this allllll by hand. We don't want to depend on the real ORM code
    # First, add a base record
    r = conn.execute(
        sa.text(
            """INSERT INTO base_record (record_type, is_service, status, created_on, modified_on)
               VALUES ('reaction', true, :status, :created_on, :modified_on)
               RETURNING id"""
        ),
        parameters=dict(status="waiting", created_on=created_on, modified_on=modified_on),
    )

    reaction_id = r.scalar()

    # Add a helpful comment
    conn.execute(
        sa.text(
            """INSERT INTO record_comment (record_id, timestamp, comment)
               VALUES (:record_id, :timestamp, :comment)"""
        ),
        parameters=dict(
            record_id=reaction_id,
            timestamp=datetime.now(timezone.utc),
            comment=f"Reaction is imported/migrated from dataset {ds_id}",
        ),
    )

    # Insert the compute history (always, for a service)
    # Fake a provenance
    provenance = {
        "creator": "qcfractal",
        "version": qcfractal_version,
        "routine": "qcfractal.services.reaction",
    }
    conn.execute(
        sa.text(
            """INSERT INTO record_compute_history (record_id, status, modified_on, provenance)
               VALUES (:record_id, :status, :modified_on, :provenance) RETURNING id"""
        ),
        parameters=dict(
            record_id=reaction_id, status="waiting", modified_on=modified_on, provenance=json.dumps(provenance)
        ),
    )

    # Now add to the reaction record table
    conn.execute(
        sa.text("INSERT INTO reaction_record (id, specification_id) VALUES (:reaction_id, :spec_id)"),
        parameters=dict(reaction_id=reaction_id, spec_id=spec_id),
    )

    stoich_map = {int(mid): coef for mid, coef in stoichiometries}

    # Now any found component records
    for sp_id, mid, _, _, _ in records:
        conn.execute(
            sa.text(
                """INSERT INTO reaction_component (reaction_id, molecule_id, coefficient, singlepoint_id)
                   VALUES (:reaction_id, :molecule_id, :coef, :sp_id)"""
            ),
            parameters=dict(reaction_id=reaction_id, molecule_id=mid, coef=stoich_map[mid], sp_id=sp_id),
        )

    # Add the service
    sp_ids = [x[0] for x in records]

    # guess a tag
    r = conn.execute(
        sa.text("SELECT tag FROM task_queue WHERE record_id IN :record_id LIMIT 1"),
        parameters=dict(record_id=tuple(sp_ids)),
    )
    tag = r.scalar_one_or_none()
    if tag is None:
        tag = "_rxn_migrated"

    # Add to the service queue
    r = conn.execute(
        sa.text(
            """INSERT INTO service_queue (record_id, tag, priority, created_on)
           VALUES (:reaction_id, :tag, :priority, :created_on)
           RETURNING id
        """
        ),
        parameters=dict(reaction_id=reaction_id, created_on=created_on, tag=tag, priority=1),
    )

    return reaction_id


def create_manybody_record(conn, ds_id, created_on, modified_on, spec_id, mol_id, record):
    # We do this allllll by hand. We don't want to depend on the real ORM code
    # First, add a base record
    r = conn.execute(
        sa.text(
            """INSERT INTO base_record (record_type, is_service, status, created_on, modified_on)
               VALUES ('manybody', true, :status, :created_on, :modified_on)
               RETURNING id"""
        ),
        parameters=dict(status="waiting", created_on=created_on, modified_on=modified_on),
    )

    manybody_id = r.scalar()

    # Add a helpful comment
    conn.execute(
        sa.text(
            """INSERT INTO record_comment (record_id, timestamp, comment)
               VALUES (:record_id, :timestamp, :comment)"""
        ),
        parameters=dict(
            record_id=manybody_id,
            timestamp=datetime.now(timezone.utc),
            comment=f"Manybody record is imported/migrated from reaction dataset {ds_id}",
        ),
    )

    # Insert the compute history (always, for a service)
    # Fake a provenance
    provenance = {
        "creator": "qcfractal",
        "version": qcfractal_version,
        "routine": "qcfractal.services.manybody",
    }
    conn.execute(
        sa.text(
            """INSERT INTO record_compute_history (record_id, status, modified_on, provenance)
               VALUES (:record_id, :status, :modified_on, :provenance) RETURNING id"""
        ),
        parameters=dict(
            record_id=manybody_id, status="waiting", modified_on=modified_on, provenance=json.dumps(provenance)
        ),
    )

    # Now add to the manybody record table
    conn.execute(
        sa.text(
            "INSERT INTO manybody_record (id, initial_molecule_id, specification_id) VALUES (:manybody_id, :mol_id, :spec_id)"
        ),
        parameters=dict(manybody_id=manybody_id, mol_id=mol_id, spec_id=spec_id),
    )

    # guess a tag
    r = conn.execute(
        sa.text("SELECT tag FROM task_queue WHERE record_id = :record_id LIMIT 1"), parameters={"record_id": record[0]}
    )

    tag = r.scalar_one_or_none()
    if tag is None:
        tag = "_mb_migrated"

    # Add to the service queue
    r = conn.execute(
        sa.text(
            """INSERT INTO service_queue (record_id, tag, priority, created_on)
           VALUES (:manybody_id, :tag, :priority, :created_on)
           RETURNING id
        """
        ),
        parameters=dict(manybody_id=manybody_id, created_on=created_on, tag=tag, priority=1),
    )

    return manybody_id


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    #####################################
    # New specification table
    #####################################
    op.create_table(
        "reaction_dataset_specification",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["reaction_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["specification_id"],
            ["reaction_specification.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "name"),
    )
    op.create_index(
        "ix_reaction_dataset_specification_dataset_id", "reaction_dataset_specification", ["dataset_id"], unique=False
    )
    op.create_index("ix_reaction_dataset_specification_name", "reaction_dataset_specification", ["name"], unique=False)
    op.create_index(
        "ix_reaction_dataset_specification_specification_id",
        "reaction_dataset_specification",
        ["specification_id"],
        unique=False,
    )

    #####################################
    # Modify existing entry table
    #####################################
    op.alter_column("reaction_dataset_entry", "reaction_dataset_id", new_column_name="dataset_id")
    op.add_column("reaction_dataset_entry", sa.Column("comment", sa.String(), nullable=True))
    op.add_column(
        "reaction_dataset_entry",
        sa.Column("additional_keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    )
    op.alter_column(
        "reaction_dataset_entry",
        "attributes",
        existing_type=postgresql.JSON(astext_type=sa.Text()),
        type_=postgresql.JSONB(astext_type=sa.Text()),
        existing_nullable=True,
    )
    op.create_index("ix_reaction_dataset_entry_dataset_id", "reaction_dataset_entry", ["dataset_id"], unique=False)
    op.create_index("ix_reaction_dataset_entry_name", "reaction_dataset_entry", ["name"], unique=False)
    op.drop_constraint("reaction_dataset_entry_reaction_dataset_id_fkey", "reaction_dataset_entry", type_="foreignkey")
    op.create_foreign_key(
        None, "reaction_dataset_entry", "reaction_dataset", ["dataset_id"], ["id"], ondelete="cascade"
    )
    op.drop_constraint("reaction_dataset_entry_pkey", "reaction_dataset_entry", type_="primary")
    op.create_primary_key("reaction_dataset_entry_pkey", "reaction_dataset_entry", ["dataset_id", "name"])

    # Attributes should be not nullable
    # Also, merge extras with attributes
    op.execute(sa.text("UPDATE reaction_dataset_entry SET attributes='{}'::json WHERE attributes IS NULL"))
    op.execute(
        sa.text("UPDATE reaction_dataset_entry SET additional_keywords='{}'::json WHERE additional_keywords IS NULL")
    )
    op.execute(sa.text("UPDATE reaction_dataset_entry SET attributes = (extras::jsonb || attributes::jsonb)"))

    #####################################
    # New stoichiometry table
    #####################################
    op.create_table(
        "reaction_dataset_stoichiometry",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("entry_name", sa.String(), nullable=False),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.Column("coefficient", postgresql.DOUBLE_PRECISION(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["reaction_dataset_entry.dataset_id", "reaction_dataset_entry.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(["dataset_id"], ["reaction_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["molecule_id"],
            ["molecule.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "entry_name", "molecule_id"),
    )
    op.create_index(
        "ix_reaction_dataset_stoichiometry_dataset_id", "reaction_dataset_stoichiometry", ["dataset_id"], unique=False
    )
    op.create_index(
        "ix_reaction_dataset_stoichiometry_entry_name", "reaction_dataset_stoichiometry", ["entry_name"], unique=False
    )
    op.create_index(
        "ix_reaction_dataset_stoichiometry_molecule_id", "reaction_dataset_stoichiometry", ["molecule_id"], unique=False
    )

    #####################################
    # New record (item) table
    #####################################
    op.create_table(
        "reaction_dataset_record",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("entry_name", sa.String(), nullable=False),
        sa.Column("specification_name", sa.String(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["reaction_dataset_entry.dataset_id", "reaction_dataset_entry.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["reaction_dataset_specification.dataset_id", "reaction_dataset_specification.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(["dataset_id"], ["reaction_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["record_id"],
            ["reaction_record.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "entry_name", "specification_name"),
        sa.UniqueConstraint("dataset_id", "entry_name", "specification_name", name="ux_reaction_dataset_record_unique"),
    )
    op.create_index("ix_reaction_dataset_record_record_id", "reaction_dataset_record", ["record_id"], unique=False)

    ######################
    # Migrate data
    ######################
    conn = op.get_bind()
    empty_kw = get_empty_keywords_id(conn)

    op.execute(sa.text("UPDATE collection SET collection = 'reaction' where collection = 'reactiondataset'"))

    # rxn -> reaction
    # ie -> manybody
    op.execute(
        sa.text(
            """UPDATE collection
                   SET collection = 'reaction', collection_type = 'reaction'
                   FROM reaction_dataset
                   WHERE collection.collection = 'reaction'
                   AND reaction_dataset.id = collection.id
                   AND reaction_dataset.ds_type = 'rxn'"""
        )
    )

    op.execute(
        sa.text(
            """UPDATE collection
                   SET collection = 'manybody', collection_type = 'manybody'
                   FROM reaction_dataset
                   WHERE collection.collection = 'reaction'
                   AND reaction_dataset.id = collection.id
                   AND reaction_dataset.ds_type = 'ie'"""
        )
    )

    # Existing tables
    rxn_dataset_table = table(
        "reaction_dataset",
        column("id", sa.Integer),
        column("ds_type", sa.String),
        column("history", sa.JSON),
        column("history_keys", sa.JSON),
        column("alias_keywords", sa.JSON),
    )

    rxn_entry_table = table(
        "reaction_dataset_entry",
        column("name", sa.String),
        column("dataset_id", sa.Integer),
        column("attributes", sa.Integer),
        column("reaction_results", sa.JSON),
        column("stoichiometry", sa.JSON),
    )

    session = Session(conn)
    datasets = session.query(rxn_dataset_table).all()

    for ds in datasets:
        ds_type = ds.ds_type
        entries = session.query(rxn_entry_table).filter(rxn_entry_table.c.dataset_id == ds.id).all()

        # Test before ding migrations
        # Find all stoichiometries, and bail if there are more than 1 and they have different molecules
        for entry in entries:
            stoichs = list(entry.stoichiometry.items())
            if len(stoichs) == 0:
                continue

            if ds_type == "ie":
                # Only include stoichiometries if they don't end in a number...
                stoichs = [x for x in stoichs if x[0].rstrip("0123456789") == x[0]]

            if not all(x[1] == stoichs[0][1] for x in stoichs):
                raise RuntimeError("Multiple named stoichiometries with different molecules. Wasn't planning on this")

        # If an IE calculation, create an MB dataset
        if ds_type == "ie":
            # We have to add the manybody dataset
            conn.execute(sa.text("INSERT INTO manybody_dataset (id) VALUES (:ds_id)"), parameters=dict(ds_id=ds.id))

        # Loop over the history, forming specifications
        for idx, h in enumerate(ds.history):
            spec_name = f"spec_{idx}"
            spec = dict(zip(ds.history_keys, h))

            # keywords should be in alias_keywords, except for dftd3 directly run through the
            # composition planner......
            try:
                if spec["keywords"] is None:
                    kw = None
                else:
                    kw = ds.alias_keywords[spec["program"]][spec["keywords"]]
            except KeyError:
                if spec["program"] == "dftd3":
                    kw = empty_kw
                else:
                    raise RuntimeError(f"Missing entry from alias_keywords: {spec['program']}, {spec['keywords']}")

            if kw is None:
                kw = empty_kw
            if spec["basis"] is None:
                spec["basis"] = ""

            qc_spec_id = add_qc_spec(conn, spec["program"], spec["driver"], spec["method"], spec["basis"], kw, {})

            # For reactions, we modify in-place in the reaction dataset table
            if ds_type == "rxn":
                # Add the reaction spec
                conn.execute(
                    sa.text(
                        """INSERT INTO reaction_specification (program, singlepoint_specification_id, optimization_specification_id, keywords)
                       VALUES ('reaction', :spec_id, NULL, '{}'::jsonb)
                       ON CONFLICT DO NOTHING"""
                    ),
                    parameters=dict(spec_id=qc_spec_id),
                )

                rxn_spec_res = conn.execute(
                    sa.text(
                        """
                            SELECT id FROM reaction_specification
                            WHERE program = 'reaction'
                            AND singlepoint_specification_id = :spec_id
                            AND optimization_specification_id IS NULL
                            AND keywords = '{}'::jsonb"""
                    ),
                    parameters=dict(spec_id=qc_spec_id),
                )

                rxn_spec_id = rxn_spec_res.scalar()

                # Now add to reaction dataset specification
                conn.execute(
                    sa.text(
                        """INSERT INTO reaction_dataset_specification (dataset_id, name, specification_id)
                                VALUES (:ds_id, :spec_name, :spec_id)"""
                    ),
                    parameters=dict(ds_id=ds.id, spec_name=spec_name, spec_id=rxn_spec_id),
                )

                for entry in entries:
                    stoich_for_spec = entry.stoichiometry[spec["stoichiometry"]]

                    # molecule:coefficient pairs
                    stoich_mol_coeff = list(stoich_for_spec.items())

                    # Add these to the stoichiometry table
                    for mol_id, coeff in stoich_mol_coeff:
                        conn.execute(
                            sa.text(
                                """
                            INSERT INTO reaction_dataset_stoichiometry (dataset_id, entry_name, molecule_id, coefficient)
                            VALUES (:ds_id, :name, :mol_id, :coeff)
                            ON CONFLICT DO NOTHING
                        """
                            ),
                            parameters=dict(ds_id=ds.id, name=entry.name, mol_id=mol_id, coeff=coeff),
                        )

                    # This stoichiometry has molecule ids (as str) as keys
                    mol_ids = list(int(x) for x in stoich_for_spec.keys())

                    # Get info about all the singlepoint records making up this stoichiometry
                    r = conn.execute(
                        sa.text(
                            """SELECT sp.id,sp.molecule_id,br.status,br.created_on,br.modified_on
                                                FROM singlepoint_record sp
                                                INNER JOIN base_record br on sp.id = br.id
                                                WHERE sp.specification_id = :spec_id
                                                AND sp.molecule_id IN :mol_ids
                                                """
                        ),
                        parameters=dict(spec_id=qc_spec_id, mol_ids=tuple(mol_ids)),
                    )

                    records = r.all()
                    found_mol_id = [x[1] for x in records]
                    all_status = [x[2] for x in records]

                    # Were all molecules found?
                    all_found = set(found_mol_id) == set(mol_ids)

                    if len(records) > 0:
                        # The user must have tried to compute these?
                        created_on = min(x[3] for x in records)
                        modified_on = max(x[4] for x in records)
                        rxn_id = create_reaction_record(
                            conn, ds.id, created_on, modified_on, rxn_spec_id, stoich_mol_coeff, records
                        )

                        # Now add to dataset
                        conn.execute(
                            sa.text(
                                """
                            INSERT INTO reaction_dataset_record (dataset_id, entry_name, specification_name, record_id)
                            VALUES (:ds_id, :entry_name, :spec_name, :record_id)"""
                            ),
                            parameters=dict(ds_id=ds.id, entry_name=entry.name, spec_name=spec_name, record_id=rxn_id),
                        )

            else:  # Is interaction energy (ie)
                # What type of interaction energy is this?
                ie_type = spec["stoichiometry"]

                if ie_type == "cp":
                    mb_kw = {"bsse_correction": "cp"}
                else:
                    mb_kw = {"bsse_correction": "none"}

                mb_kw_str = json.dumps(mb_kw)

                # Add the manybody spec
                conn.execute(
                    sa.text(
                        """INSERT INTO manybody_specification (program, singlepoint_specification_id, keywords)
                       VALUES ('manybody', :spec_id, :mb_kw)
                       ON CONFLICT DO NOTHING"""
                    ),
                    parameters=dict(spec_id=qc_spec_id, mb_kw=mb_kw_str),
                )

                mb_spec_res = conn.execute(
                    sa.text(
                        """
                            SELECT id FROM manybody_specification
                            WHERE program = 'manybody'
                            AND singlepoint_specification_id = :spec_id
                            AND keywords = :mb_kw"""
                    ),
                    parameters=dict(spec_id=qc_spec_id, mb_kw=mb_kw_str),
                )

                mb_spec_id = mb_spec_res.scalar()

                # Now add to manybody dataset specification
                conn.execute(
                    sa.text(
                        """INSERT INTO manybody_dataset_specification (dataset_id, name, specification_id)
                                VALUES (:ds_id, :spec_name, :spec_id)"""
                    ),
                    parameters=dict(ds_id=ds.id, spec_name=spec_name, spec_id=mb_spec_id),
                )

                for entry in entries:
                    # we only want the main molecule ("stoichiometry" key that is without a number)
                    stoich_for_spec = entry.stoichiometry[spec["stoichiometry"]]

                    # Should only be one molecule
                    assert len(stoich_for_spec) == 1

                    mol_id = int(list(stoich_for_spec.keys())[0])

                    # Add the entry
                    conn.execute(
                        sa.text(
                            """
                        INSERT INTO manybody_dataset_entry (dataset_id, name, initial_molecule_id, attributes, additional_keywords)
                        VALUES (:ds_id, :name, :mol_id, :attrib, '{}'::jsonb)
                        ON CONFLICT DO NOTHING
                        """
                        ),
                        parameters=dict(
                            ds_id=ds.id, name=entry.name, mol_id=mol_id, attrib=json.dumps(entry.attributes)
                        ),
                    )

                    # See if this parent molecule has been computed
                    # (so we can get created_on, etc)
                    r = conn.execute(
                        sa.text(
                            """SELECT sp.id,sp.molecule_id,br.status,br.created_on,br.modified_on
                               FROM singlepoint_record sp
                               INNER JOIN base_record br on sp.id = br.id
                               WHERE sp.specification_id = :spec_id
                               AND sp.molecule_id = :mol_id
                               """
                        ),
                        parameters=dict(spec_id=qc_spec_id, mol_id=mol_id),
                    )

                    record = r.one_or_none()

                    # Just add an MBE service with this molecule, and let it
                    # find all the fragments

                    if record is not None:
                        # The user must have tried to compute these?
                        created_on = record[3]
                        modified_on = record[4]
                        mb_id = create_manybody_record(conn, ds.id, created_on, modified_on, mb_spec_id, mol_id, record)

                        # Now add to dataset
                        conn.execute(
                            sa.text(
                                """
                            INSERT INTO manybody_dataset_record (dataset_id, entry_name, specification_name, record_id)
                            VALUES (:ds_id, :entry_name, :spec_name, :record_id)"""
                            ),
                            parameters=dict(ds_id=ds.id, entry_name=entry.name, spec_name=spec_name, record_id=mb_id),
                        )

                # Remove from the reaction dataset table
                conn.execute(sa.text("DELETE FROM reaction_dataset WHERE id = :ds_id"), parameters=dict(ds_id=ds.id))

    #########################################
    # Modify existing reaction_dataset table
    #########################################

    # Make columns not nullable
    op.alter_column("reaction_dataset_entry", "attributes", nullable=False)

    # Drop unused columns
    op.drop_column("reaction_dataset_entry", "extras")
    op.drop_column("reaction_dataset_entry", "reaction_results")
    op.drop_column("reaction_dataset_entry", "stoichiometry")
    op.drop_column("reaction_dataset", "default_program")
    op.drop_column("reaction_dataset", "default_driver")
    op.drop_column("reaction_dataset", "alias_keywords")
    op.drop_column("reaction_dataset", "default_keywords")
    op.drop_column("reaction_dataset", "ds_type")
    op.drop_column("reaction_dataset", "history_keys")
    op.drop_column("reaction_dataset", "history")
    op.drop_column("reaction_dataset", "default_units")
    op.drop_column("reaction_dataset", "default_benchmark")

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""finalize migration of outputs

Revision ID: c23b3ba12869
Revises: 5a98e27e4c10
Create Date: 2023-03-03 14:56:36.329930

"""

import json
import lzma

import sqlalchemy as sa
import zstandard
from alembic import op
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcportal.compression import compress, CompressionEnum

# revision identifiers, used by Alembic.
revision = "c23b3ba12869"
down_revision = "5a98e27e4c10"
branch_labels = None
depends_on = None


def decompress_old_string(compressed_data: bytes, compression_type: CompressionEnum) -> str:
    if compression_type == CompressionEnum.none:
        r = compressed_data
    elif compression_type == CompressionEnum.lzma:
        r = lzma.decompress(compressed_data)
    elif compression_type == CompressionEnum.zstd:
        r = zstandard.decompress(compressed_data)
    else:
        # Shouldn't ever happen, unless we change CompressionEnum but not the rest of this function
        raise TypeError(f"Unknown compression type: {compression_type}")
    return r.decode()


def upgrade():
    # Table with output data
    old_output_table = table(
        "output_store",
        column("id", sa.Integer),
        column("compression_type", sa.Enum(CompressionEnum)),
        column("compression_level", sa.Integer),
        column("data", sa.LargeBinary),
        column("value", sa.JSON),
        column("old_data", sa.LargeBinary),
    )

    conn = op.get_bind()
    session = Session(conn)

    old_outs = session.query(old_output_table).yield_per(10)

    for old_out in old_outs:
        if old_out.old_data is not None:
            old_data = decompress_old_string(old_out.old_data, old_out.compression_type)
            if old_data.startswith("{") and "error_type" in old_data:
                old_data = json.loads(old_data)
            new_data, ctype, clevel = compress(old_data, CompressionEnum.zstd)
        elif old_out.data is None:
            assert old_out.compression_type is None or old_out.compression_type == "none"
            new_data, ctype, clevel = compress(old_out.value, CompressionEnum.zstd)
        else:
            new_data = None

        if new_data is not None:
            conn.execute(
                sa.text(
                    """UPDATE output_store SET
                         value=NULL,
                         old_data=NULL,
                         compression_type=:ctype,
                         compression_level=:clevel,
                         data=:cdata
                       WHERE id = :id"""
                ),
                parameters={"id": old_out.id, "cdata": new_data, "ctype": ctype, "clevel": clevel},
            )

    op.drop_column("output_store", "old_data")
    op.drop_column("output_store", "value")
    op.alter_column("output_store", "data", nullable=False)
    op.alter_column("output_store", "compression_level", nullable=False)
    op.alter_column("output_store", "compression_type", nullable=False)


def downgrade():
    raise RuntimeError("Cannot downgrade")
    pass



"""migrate service dependencies table

Revision ID: 31651dcef18d
Revises: 7a60e93aed72
Create Date: 2021-12-22 15:29:29.396561

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "31651dcef18d"
down_revision = "7a60e93aed72"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table("service_queue_tasks", "service_dependencies")

    op.alter_column(
        "service_dependencies",
        "extras",
        existing_type=postgresql.JSON(astext_type=sa.Text()),
        type_=postgresql.JSONB(astext_type=sa.Text()),
        nullable=False,
    )
    op.create_unique_constraint(
        "ux_service_dependencies", "service_dependencies", ["service_id", "record_id", "extras"]
    )

    op.execute(sa.text("ALTER SEQUENCE service_queue_tasks_id_seq RENAME TO service_dependencies_id_seq"))
    op.execute(sa.text("ALTER INDEX service_queue_tasks_pkey RENAME TO service_dependencies_pkey"))
    op.execute(
        sa.text(
            "ALTER TABLE service_dependencies RENAME CONSTRAINT service_queue_tasks_record_id_fkey TO service_dependencies_record_id_fkey"
        )
    )
    op.execute(
        sa.text(
            "ALTER TABLE service_dependencies RENAME CONSTRAINT service_queue_tasks_service_id_fkey TO service_dependencies_service_id_fkey"
        )
    )

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###



"""add neb nextchain subtask

Revision ID: 8c781020aed4
Revises: ee6defa818d6
Create Date: 2023-01-19 20:48:18.953059

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8c781020aed4"
down_revision = "ee6defa818d6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "neb_subtasks",
        sa.Column("neb_id", sa.Integer(), nullable=False),
        sa.Column("subtask_id", sa.Integer(), nullable=False),
        sa.Column("chain_iteration", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["neb_id"], ["neb_record.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["subtask_id"],
            ["service_subtask_record.id"],
        ),
        sa.PrimaryKeyConstraint("neb_id", "subtask_id", "chain_iteration"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("neb_subtasks")
    # ### end Alembic commands ###



"""Migrate torsiondrive datasets

Revision ID: bea426877db0
Revises: efaea8a3b4c0
Create Date: 2022-03-08 08:48:47.605268

"""

import json
import os
import sys

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers.v0_50_helpers import get_empty_keywords_id, add_opt_spec, add_qc_spec

# revision identifiers, used by Alembic.
revision = "bea426877db0"
down_revision = "efaea8a3b4c0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "torsiondrive_dataset",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["collection.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "torsiondrive_dataset_entry",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("comment", sa.String(), nullable=True),
        sa.Column("torsiondrive_keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("additional_keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column("attributes", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.ForeignKeyConstraint(["dataset_id"], ["torsiondrive_dataset.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("dataset_id", "name"),
    )
    op.create_index(
        "ix_torsiondrive_dataset_entry_dataset_id", "torsiondrive_dataset_entry", ["dataset_id"], unique=False
    )

    op.create_index("ix_torsiondrive_dataset_entry_name", "torsiondrive_dataset_entry", ["name"], unique=False)

    op.create_table(
        "torsiondrive_dataset_molecule",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("entry_name", sa.String(), nullable=False),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["torsiondrive_dataset_entry.dataset_id", "torsiondrive_dataset_entry.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["molecule_id"],
            ["molecule.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "entry_name", "molecule_id"),
    )

    op.create_index(
        "ix_torsiondrive_dataset_molecule_dataset_id",
        "torsiondrive_dataset_molecule",
        ["dataset_id"],
        unique=False,
    )

    op.create_index(
        "ix_torsiondrive_dataset_molecule_entry_name",
        "torsiondrive_dataset_molecule",
        ["entry_name"],
        unique=False,
    )

    op.create_index(
        "ix_torsiondrive_dataset_molecule_molecule_id",
        "torsiondrive_dataset_molecule",
        ["molecule_id"],
        unique=False,
    )

    op.create_table(
        "torsiondrive_dataset_specification",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["torsiondrive_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["specification_id"],
            ["optimization_specification.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "name"),
    )
    op.create_index(
        "ix_torsiondrive_dataset_specification_dataset_id",
        "torsiondrive_dataset_specification",
        ["dataset_id"],
        unique=False,
    )
    op.create_index(
        "ix_torsiondrive_dataset_specification_name", "torsiondrive_dataset_specification", ["name"], unique=False
    )
    op.create_index(
        "ix_torsiondrive_dataset_specification_specification_id",
        "torsiondrive_dataset_specification",
        ["specification_id"],
        unique=False,
    )

    op.create_table(
        "torsiondrive_dataset_record",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("entry_name", sa.String(), nullable=False),
        sa.Column("specification_name", sa.String(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["torsiondrive_dataset_entry.dataset_id", "torsiondrive_dataset_entry.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["torsiondrive_dataset_specification.dataset_id", "torsiondrive_dataset_specification.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(["dataset_id"], ["torsiondrive_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["record_id"],
            ["torsiondrive_record.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "entry_name", "specification_name"),
        sa.UniqueConstraint(
            "dataset_id", "entry_name", "specification_name", name="ux_torsiondrive_dataset_record_unique"
        ),
    )
    op.create_index(
        "ix_torsiondrive_dataset_record_record_id", "torsiondrive_dataset_record", ["record_id"], unique=False
    )

    ######################
    # MIGRATION
    ######################
    conn = op.get_bind()

    op.execute(
        sa.text("UPDATE collection SET collection_type = 'torsiondrive' where collection = 'torsiondrivedataset'")
    )

    op.execute(sa.text("UPDATE collection SET collection = 'torsiondrive' where collection = 'torsiondrivedataset'"))

    # Temporary ORM
    collection_table = table(
        "collection",
        column("id", sa.Integer),
        column("collection", sa.String),
        column("extra", sa.JSON),
    )

    session = Session(conn)
    collections = session.query(collection_table).where(collection_table.c.collection == "torsiondrive").all()
    for col in collections:
        ext = col.extra
        col_specs = ext.pop("specs")
        col_record = ext.pop("records")

        # Add the dataset to the separate optimziation dataset table
        conn.execute(sa.text("INSERT INTO torsiondrive_dataset (id) VALUES (:colid)"), parameters={"colid": col.id})

        # Specifications
        # Empty keywords
        empty_kw_id = get_empty_keywords_id(conn)

        #########################
        # SPECIFICATIONS
        #########################
        for spec_name, spec in col_specs.items():
            # Protocols for qc_spec were always ignored. So set them with the default
            opt_spec = spec["optimization_spec"]
            qc_spec = spec["qc_spec"]
            qc_spec["driver"] = "deferred"

            if qc_spec.get("keywords", None) is None:
                qc_spec["keywords"] = empty_kw_id

            qc_spec_id = add_qc_spec(
                conn,
                qc_spec["program"],
                qc_spec["driver"],
                qc_spec["method"],
                qc_spec["basis"],
                qc_spec["keywords"],
                {},
            )

            # Optimization spec
            if opt_spec.get("protocols", None) is None:
                opt_spec["protocols"] = {}
            if opt_spec.get("keywords", None) is None:
                opt_spec["keywords"] = {}

            opt_spec_id = add_opt_spec(
                conn, qc_spec_id, opt_spec["program"], opt_spec["keywords"], opt_spec["protocols"]
            )

            # Now add to the dataset spec
            conn.execute(
                sa.text(
                    """
                       INSERT INTO torsiondrive_dataset_specification (dataset_id, name, description, specification_id)
                       VALUES (:col_id, :spec_name, :spec_desc, :opt_spec_id)
                    """
                ),
                parameters=dict(
                    col_id=col.id,
                    spec_name=spec["name"],
                    spec_desc=spec["description"],
                    opt_spec_id=opt_spec_id,
                ),
            )

        ####################
        # ENTRIES/RECORDS
        ####################
        conn.execute(
            sa.text(
                """
                   INSERT INTO torsiondrive_dataset_entry (dataset_id, name, torsiondrive_keywords, additional_keywords, attributes)
                   SELECT :col_id, j.key, (j.value->>'td_keywords')::jsonb, (j.value->'td_keywords'->'additional_keywords')::jsonb, (j.value->>'attributes')::jsonb
                   FROM collection, json_each(extra->'records') AS j
                   WHERE collection.id = :col_id
                """
            ),
            parameters={"col_id": col.id},
        )

        conn.execute(
            sa.text(
                """
                   UPDATE torsiondrive_dataset_entry
                   SET torsiondrive_keywords = torsiondrive_keywords - 'additional_keywords'
                   WHERE torsiondrive_dataset_entry.dataset_id = :col_id
                """
            ),
            parameters={"col_id": col.id},
        )

        conn.execute(
            sa.text(
                """
                    WITH cte AS (
                        SELECT j.key as entry_name, json_array_elements_text(j.value->'initial_molecules')::integer as mol_id
                        FROM collection, json_each(extra->'records') AS j
                        WHERE collection.id = :col_id)
                   INSERT INTO torsiondrive_dataset_molecule (dataset_id, entry_name, molecule_id)
                   SELECT :col_id, entry_name, molecule.id FROM cte, molecule
                   WHERE molecule.id = cte.mol_id
                   """
            ),
            parameters={"col_id": col.id},
        )

        for ent_name, ent in col_record.items():
            for spec_name, record_id in ent["object_map"].items():
                conn.execute(
                    sa.text(
                        """
                           INSERT INTO torsiondrive_dataset_record (dataset_id, entry_name, specification_name, record_id)
                           SELECT :col_id, :ent_name, :spec_name, r.id
                           FROM torsiondrive_record r
                           WHERE r.id = :record_id
                           """
                    ),
                    parameters=dict(col_id=col.id, ent_name=ent_name, spec_name=spec_name, record_id=record_id),
                )

        # Update the collection extra, with the removed fields
        ext.pop("history", None)
        conn.execute(
            sa.text("UPDATE collection SET extra = (:extra)::json WHERE id = :col_id"),
            parameters=dict(col_id=col.id, extra=json.dumps(ext)),
        )

    # ### end Alembic commands ###


def downgrade():
    raise RuntimeError("Cannot downgrade")



"""refactor torsiondrive

Revision ID: bb4804242cea
Revises: 31651dcef18d
Create Date: 2021-12-17 10:23:34.593275

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import table, column

from qcfractal.db_socket.column_types import PlainMsgpackExt

# revision identifiers, used by Alembic.
revision = "bb4804242cea"
down_revision = "31651dcef18d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # Torsiondrive spec table
    op.create_table(
        "torsiondrive_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(length=100), nullable=False),
        sa.Column("optimization_specification_id", sa.Integer(), nullable=False),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_torsiondrive_specification_program_lower"),
        sa.ForeignKeyConstraint(
            ["optimization_specification_id"],
            ["optimization_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "program", "optimization_specification_id", "keywords", name="ux_torsiondrive_specification_keys"
        ),
    )

    op.create_index("ix_torsiondrive_specification_keywords", "torsiondrive_specification", ["keywords"], unique=False)
    op.create_index(
        "ix_torsiondrive_specification_optimization_specification_id",
        "torsiondrive_specification",
        ["optimization_specification_id"],
        unique=False,
    )
    op.create_index("ix_torsiondrive_specification_program", "torsiondrive_specification", ["program"], unique=False)

    # Modify the optimization history table
    op.alter_column("optimization_history", "opt_id", new_column_name="optimization_id")
    op.alter_column("optimization_history", "torsion_id", new_column_name="torsiondrive_id")
    op.drop_constraint("optimization_history_torsion_id_fkey", "optimization_history", type_="foreignkey")
    op.drop_constraint("optimization_history_opt_id_fkey", "optimization_history", type_="foreignkey")
    op.create_foreign_key(
        "torsiondrive_optimizations_torsiondrive_id_fkey",
        "optimization_history",
        "torsiondrive_procedure",
        ["torsiondrive_id"],
        ["id"],
        ondelete="cascade",
    )
    op.create_foreign_key(
        "torsiondrive_optimizations_optimization_id_fkey",
        "optimization_history",
        "optimization_record",
        ["optimization_id"],
        ["id"],
    )

    op.drop_constraint("optimization_history_pkey", "optimization_history", type_="primary")
    op.create_primary_key(
        "torsiondrive_optimizations_pkey",
        "optimization_history",
        ["torsiondrive_id", "optimization_id", "key", "position"],
    )

    # Initial molecule association table
    op.alter_column("torsion_init_mol_association", "torsion_id", new_column_name="torsiondrive_id")
    op.drop_constraint(
        "torsion_init_mol_association_torsion_id_fkey", "torsion_init_mol_association", type_="foreignkey"
    )
    op.drop_constraint(
        "torsion_init_mol_association_molecule_id_fkey", "torsion_init_mol_association", type_="foreignkey"
    )
    op.create_foreign_key(
        "torsiondrive_initial_molecules_torsiondrive_id_fkey",
        "torsion_init_mol_association",
        "torsiondrive_procedure",
        ["torsiondrive_id"],
        ["id"],
        ondelete="cascade",
    )
    op.create_foreign_key(
        "torsiondrive_initial_molecules_molecule_id_fkey",
        "torsion_init_mol_association",
        "molecule",
        ["molecule_id"],
        ["id"],
    )
    op.drop_constraint("torsion_init_mol_association_pkey", "torsion_init_mol_association", type_="primary")
    op.create_primary_key(
        "torsiondrive_initial_molecules_pkey", "torsion_init_mol_association", ["torsiondrive_id", "molecule_id"]
    )

    # The torsiondrive record table itself
    op.add_column("torsiondrive_procedure", sa.Column("specification_id", sa.Integer(), nullable=True))
    op.create_foreign_key(
        "torsiondrive_record_specification_id_fkey",
        "torsiondrive_procedure",
        "torsiondrive_specification",
        ["specification_id"],
        ["id"],
    )

    op.execute(sa.text("ALTER INDEX torsiondrive_procedure_pkey RENAME TO torsiondrive_record_pkey"))
    op.execute(
        sa.text(
            "ALTER TABLE torsiondrive_procedure RENAME CONSTRAINT torsiondrive_procedure_id_fkey TO torsiondrive_record_id_fkey"
        )
    )

    ###########################################################
    # NOW THE BIG MIGRATION
    ###########################################################
    # change the spec columns to jsonb
    op.execute("ALTER TABLE torsiondrive_procedure ALTER COLUMN qc_spec TYPE JSONB")
    op.execute("ALTER TABLE torsiondrive_procedure ALTER COLUMN optimization_spec TYPE JSONB")
    op.execute("ALTER TABLE torsiondrive_procedure ALTER COLUMN keywords TYPE JSONB")

    # Update the optimization spec with the td keywords additional keywords
    op.execute(
        sa.text(
            r"""UPDATE torsiondrive_procedure
                SET optimization_spec = optimization_spec || jsonb_build_object('keywords', (keywords->'additional_keywords') || (optimization_spec->'keywords'))
                WHERE keywords ? 'additional_keywords'"""
        )
    )

    # Remove additional keywords from TD keywords
    op.execute(
        sa.text(
            """UPDATE torsiondrive_procedure
               SET keywords = keywords - 'additional_keywords'"""
        )
    )

    # The empty, default keywords
    res = op.get_bind().execute(
        sa.text("SELECT id FROM keywords WHERE hash_index = 'bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f'")
    )
    empty_kw = res.scalar()

    # Fiddle with the specifications
    # First, a hack. The MolSSI database has some old data with a non-existent hash still there. Replace
    # that with the appropriate keyword
    op.execute(
        sa.text(
            r"""UPDATE torsiondrive_procedure SET qc_spec = (qc_spec || '{"keywords": "2"}') WHERE qc_spec->>'keywords' = '5c954fa6b6a2de5f188ea234'"""
        )
    )

    # Remove empty and null keywords
    op.execute(
        sa.text(r"UPDATE torsiondrive_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' = ''")
    )
    op.execute(
        sa.text(
            r"UPDATE torsiondrive_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' = 'null'"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE torsiondrive_procedure SET qc_spec = (qc_spec - 'keywords') WHERE qc_spec->>'keywords' IS NULL"
        )
    )

    # Insert the qcspec
    # Protocols for qc_spec were always ignored. So set them with the default
    op.execute(
        sa.text(
            f"""
               INSERT INTO qc_specification (program, driver, method, basis, keywords_id, protocols)
               SELECT DISTINCT td.qc_spec->>'program',
                               'deferred'::singlepointdriver,
                               td.qc_spec->>'method',
                               COALESCE(td.qc_spec->>'basis', ''),
                               COALESCE((td.qc_spec->>'keywords')::int, {empty_kw}),
                               '{{}}'::jsonb
               FROM torsiondrive_procedure td
               ON CONFLICT DO NOTHING
               """
        )
    )

    # remove explicitly specified default protocols
    op.execute(
        sa.text(
            r"""UPDATE torsiondrive_procedure SET optimization_spec = (optimization_spec || '{"protocols": "{}"}') WHERE optimization_spec->'protocols' = '{"trajectory": "all"}'"""
        )
    )

    # Now the optimization_spec
    # Remove empty and null keywords
    op.execute(
        sa.text(
            r"UPDATE torsiondrive_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' = ''"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE torsiondrive_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' IS NULL"
        )
    )
    op.execute(
        sa.text(
            r"UPDATE torsiondrive_procedure SET optimization_spec = (optimization_spec - 'keywords') WHERE optimization_spec->>'keywords' = 'null'"
        )
    )

    op.execute(
        sa.text(
            f"""
               INSERT INTO optimization_specification (program, keywords, protocols, qc_specification_id)
               SELECT DISTINCT td.optimization_spec->>'program',
                               COALESCE(td.optimization_spec->'keywords', '{{}}'),
                               COALESCE(td.optimization_spec->'protocols', '{{}}'),
                               (
                               SELECT id from qc_specification sp
                               WHERE sp.program = td.qc_spec->>'program'
                               AND sp.driver = 'deferred'::singlepointdriver
                               AND sp.method = td.qc_spec->>'method'
                               AND sp.basis = COALESCE(td.qc_spec->>'basis', '')
                               AND sp.keywords_id = COALESCE((td.qc_spec->>'keywords')::int, {empty_kw})
                               AND sp.protocols = '{{}}'
                               )
               FROM torsiondrive_procedure td
               ON CONFLICT DO NOTHING
               """
        )
    )

    # And the torsiondrive spec
    op.execute(
        sa.text(
            f"""
               INSERT INTO torsiondrive_specification (program, keywords, optimization_specification_id)
               SELECT DISTINCT 'torsiondrive',
                               td.keywords,
                               (
                                    SELECT id from optimization_specification os
                                    WHERE os.program = td.optimization_spec->>'program'
                                    AND os.keywords = COALESCE(td.optimization_spec->'keywords', '{{}}')
                                    AND os.protocols = COALESCE(td.optimization_spec->'protocols', '{{}}')
                                    AND os.qc_specification_id =
                                       (
                                           SELECT id from qc_specification sp
                                           WHERE sp.program = td.qc_spec->>'program'
                                           AND sp.driver = 'deferred'::singlepointdriver
                                           AND sp.method = td.qc_spec->>'method'
                                           AND sp.basis = COALESCE(td.qc_spec->>'basis', '')
                                           AND sp.keywords_id = COALESCE((td.qc_spec->>'keywords')::int, {empty_kw})
                                           AND sp.protocols = '{{}}'
                                       )
                                )
               FROM torsiondrive_procedure td
               """
        )
    )

    # Now add this to the torsiondrive spec column
    op.execute(
        sa.text(
            f"""
               UPDATE torsiondrive_procedure td
               SET specification_id = (
                   SELECT id FROM torsiondrive_specification ts
                   WHERE ts.program = 'torsiondrive'
                   AND ts.keywords = td.keywords
                   AND ts.optimization_specification_id = (
                             SELECT id from optimization_specification os
                             WHERE os.program = td.optimization_spec->>'program'
                             AND os.keywords = COALESCE(td.optimization_spec->'keywords', '{{}}')
                             AND os.protocols = COALESCE(td.optimization_spec->'protocols', '{{}}')
                             AND os.qc_specification_id = (
                                    SELECT id from qc_specification sp
                                    WHERE sp.program = td.qc_spec->>'program'
                                    AND sp.driver = 'deferred'::singlepointdriver
                                    AND sp.method = td.qc_spec->>'method'
                                    AND sp.basis = COALESCE(td.qc_spec->>'basis', '')
                                    AND sp.keywords_id = COALESCE((td.qc_spec->>'keywords')::int, {empty_kw})
                                    AND sp.protocols = '{{}}'
                            )
                        )
                   )
               """
        )
    )

    # Now migrate the service queue
    # Temporary ORM
    service_table = table(
        "service_queue",
        column("id", sa.Integer),
        column("service_state", PlainMsgpackExt),
    )

    bind = op.get_bind()
    session = Session(bind=bind)

    services = session.query(service_table).all()

    for service in services:
        if "torsiondrive_state" not in service.service_state:
            continue

        # We have a torsiondrive
        # Remove the optimization template
        service.service_state.pop("optimization_template")
        session.execute(
            service_table.update()
            .where(service_table.c.id == service.id)
            .values({"service_state": service.service_state})
        )

    # drop the final energies and minimum position columns
    op.drop_column("torsiondrive_procedure", "final_energy_dict")
    op.drop_column("torsiondrive_procedure", "minimum_positions")

    # Make columns not nullable now that they are populated
    op.alter_column("torsiondrive_procedure", "specification_id", nullable=False)

    op.drop_column("torsiondrive_procedure", "keywords")
    op.drop_column("torsiondrive_procedure", "optimization_spec")
    op.drop_column("torsiondrive_procedure", "qc_spec")
    op.rename_table("optimization_history", "torsiondrive_optimizations")
    op.rename_table("torsiondrive_procedure", "torsiondrive_record")
    op.rename_table("torsion_init_mol_association", "torsiondrive_initial_molecules")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise RuntimeError("Cannot downgrade")



import time

import pytest

from qcfractal.snowflake import FractalSnowflake


@pytest.mark.slow
def test_snowflake_restarting():
    s = FractalSnowflake(start=False)

    s._start_api()
    s._start_compute()
    s._start_job_runner()

    time.sleep(5)

    s.stop()

    time.sleep(5)

    assert s._api_proc is None
    assert s._compute_proc is None
    assert s._job_runner_proc is None

    s._start_api()
    s._start_compute()
    s._start_job_runner()

    assert s._api_proc.is_alive()
    assert s._compute_proc.is_alive()
    assert s._job_runner_proc.is_alive()

    c = s.client()
    assert c.ping()



"""add record comments

Revision ID: bf31e6366a10
Revises: 1ebfc89accb9
Create Date: 2021-12-17 16:54:02.355727

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "bf31e6366a10"
down_revision = "1ebfc89accb9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "record_comments",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=False),
        sa.Column("username", sa.String(), nullable=True),
        sa.Column("comment", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(["record_id"], ["base_record.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("record_comments")
    # ### end Alembic commands ###
